﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Net;
using System.Net.Http;
using System.Security.Cryptography;
using System.ServiceModel;
using System.Text;
using System.Threading;
using System.Threading.Tasks;
using System.Windows;
using System.Xml.Linq;
using UserStageExample.FlexiCapture;

namespace UserStageExample.FlexiService
{
    /// <summary>
    /// Коннектор к системе FlexiCapture
    /// </summary>
    internal class FlexiConnectorService : IDisposable
    {
        /// <summary>
        /// Путь к API FlexiCapture
        /// </summary>
        private const string FlexiHandlerPath = "/FlexiCapture12/Server/{0}API/v1/Soap";
        /// <summary>
        /// Путь к аутентификации FlexiCapture
        /// </summary>
        private const string FlexiCaptureAuthPart = "FCAuth/";
        /// <summary>
        /// Допустимый размер сообщения, передаваемого в API
        /// </summary>
        private const int MessageSize = 4*1024*1024;
        /// <summary>
        /// Название проекта, с которым работает пример
        /// </summary>
        private const string UserStageProjectName = "UserStageExample";
        /// <summary>
        /// Название сервисной стадии
        /// </summary>
        private const string UserStageName = "Service";
        
        /// <summary>
        /// Роль пользователя - оператор пользовательской станции
        /// </summary>
        private const int RoleType = 12;
        /// <summary>
        /// Тип станции - пользовательская станция
        /// </summary>
        private const int WorkstationType = 10;
        /// <summary>
        /// Тип стадии исключения
        /// </summary>
        private const int ExceptionStageType = 1000;
        /// <summary>
        /// Имя файла, содержащего распознанные данные
        /// </summary>
        private const string DocumentBodyFileName = "DocumentBody";

        /// <summary>
        /// Имя файла электронной подписи, добавляемой к подтверждённым документам
        /// </summary>
        private const string SignatureFileName = "Signature.dat";
        /// <summary>
        /// Размер электронной подписи, добавляемой к документам
        /// </summary>
        private const int SignatureSize = 1024;
        /// <summary>
        /// Тип вложения - вложение к документу
        /// </summary>
        private const int DocumentAttachment = 9;

        /// <summary>
        /// Причина отказа, добавляемая к документам
        /// </summary>
        private const string RejectReason = "Rejecting, moving to exceptions";
        /// <summary>
        /// Стадия обработки - подтверждение завершения текущего этапа, при этом пакет
        /// будет направлен дальше в соотвествии со сценарием FlexiCapture
        /// </summary>
        private const int CommitUserStageType = 410;

        private const int CloseUserStageType = -1;

        /// <summary>
        /// SOAP клиент
        /// </summary>
        private readonly FlexiCaptureWebServiceSoapClient _client;

        /// <summary>
        /// Id открытой сессии в системе FlexiCapture
        /// </summary>
        private readonly int _sessionId = 0;

        /// <summary>
        /// Id проекта в системе FlexiCapture
        /// </summary>
        private readonly int _projectId = 0;

        /// <summary>
        /// Id сервисной стадии в системе FlexiCapture
        /// </summary>
        private readonly int _serviceStageId = 0;

        /// <summary>
        /// Id стадии исключения в системе FlexiCapture
        /// </summary>
        private readonly int _exceptionsStageId = 0;

        /// <summary>
        /// Признак разрушения коннектора
        /// </summary>
        private int _disposed = 0;

        /// <summary>
        /// Генератор "электронных подписей"
        /// </summary>
        private readonly RandomNumberGenerator _rng = new RNGCryptoServiceProvider();


        /// <summary>
        /// Создание коннектора
        /// </summary>
        /// <param name="serverUri">Путь к серверу FlexiCapture</param>
        /// <param name="windowsAuth">Использовать ли Negotiate аутентификацию</param>
        /// <param name="login">Логин, игнорируется если windowsAuth == true</param>
        /// <param name="password">Пароль, игнорируется если windowsAuth == true</param>
        public FlexiConnectorService(Uri serverUri, bool windowsAuth, string login, string password)
        {
            // Создаём новую привязку
            var binding = new BasicHttpBinding();

            // Размер сообщения по-умолчанию слишком мал для FlexiCapture - увеличиваем
            binding.MaxReceivedMessageSize = MessageSize;

            // Выбираем метод защиты соединения, в зависимости от протокола 
            binding.Security.Mode = serverUri.Scheme == Uri.UriSchemeHttps
                ? BasicHttpSecurityMode.Transport
                : BasicHttpSecurityMode.TransportCredentialOnly;

            // Выбораем метод аутентификации
            binding.Security.Transport.ClientCredentialType = windowsAuth
                ? HttpClientCredentialType.Windows
                : HttpClientCredentialType.Basic;

            // Выбираем путь к API, в зависимости от метода аутентификации
            var externalSuffix = windowsAuth ? string.Empty : FlexiCaptureAuthPart;
            var remoteAddress = new EndpointAddress(new Uri(serverUri, string.Format(FlexiHandlerPath, externalSuffix)));

            // Создаем SOAP клиента
            _client = new FlexiCapture.FlexiCaptureWebServiceSoapClient(binding, remoteAddress);
            // Для аутентификации FlexiCapture указываем логин и пароль
            if (_client.ClientCredentials != null && !windowsAuth)
            {
                _client.ClientCredentials.UserName.UserName = login;
                _client.ClientCredentials.UserName.Password = password;
            }

            try
            {
                // Открываем сессию
                _sessionId = _client.OpenSession(RoleType, WorkstationType);
                var projectGuid = string.Empty;
                // Находим нужный проект
                foreach (var project in _client.GetProjects())
                {
                    if (project.Name == UserStageProjectName)
                    {
                        projectGuid = project.Guid;
                    }
                }
                if (string.IsNullOrEmpty(projectGuid))
                {
                    MessageBox.Show("На сервене нет проекта UserStageExample, необходимого для работы примера");
                }
                // Открываем проект
                _projectId = _client.OpenProject(_sessionId, projectGuid);

                // Находим и запоминаем идентификатор сервисной стадии
                var stages = _client.GetProcessingStages(_projectId, 0, 0, UserStageName);
                if (stages.Count != 1) throw new Exception("Invalid project");
                _serviceStageId = stages[0].Id;

                // Находим и запоминаем external id стадии исключений
                stages = _client.GetProcessingStages(_projectId, 0, ExceptionStageType, string.Empty);
                if (stages.Count != 1) throw new Exception("Invalid project");
                _exceptionsStageId = stages[0].ExternalId;
            }
            catch (Exception)
            {
                Dispose();
                throw;
            }
        }

        /// <summary>
        /// Возвращение списка задач, ожидающих обработки на сервисной стадии
        /// </summary>
        /// <returns></returns>
        public Dictionary<int, Task> GetTasks()
        {
            // Получаем список задач
            var tasks = _client.GetAvailableTasksByStageId(_sessionId, _projectId, 
                new FlexiCapture.stageIds() { _serviceStageId }, false);

            var result = new Dictionary<int, Task>();
            if (tasks == null || tasks.Count == 0) return result;

            foreach (var task in tasks)
            {
                // Для каждой задачи создаём объект типа Task
                var item = new Task
                {
                    Id = task.Id,
                    BatchId = task.BatchId
                };

                // Получаем список документов, входящих в задачу
                var documents = _client.GetTaskDocuments(task.Id);
                if (documents == null || documents.Count == 0) continue;
                foreach (var document in documents)
                {
                    // Загружаем сопровождающий задачу XML
                    var file = _client.LoadDocumentResult(_sessionId, document.BatchId, document.Id, DocumentBodyFileName);
                    if (file == null || file.Bytes == null) continue;
                    // Разбор XML
                    var doc = new Document(document.Id, XDocument.Load(new MemoryStream(file.Bytes)));
                    // Сохраняем разобранный список документов
                    item.Documents.Add(document.Id, doc);
                }
                result.Add(item.Id, item);
            }
            return result;
        }

        /// <summary>
        /// Запуск задачи на дальнейшую обработку
        /// </summary>
        /// <param name="batchId">Идентификатор пакета</param>
        /// <param name="taskId">Идентификатор задачи</param>
        /// <param name="acceptedDocuments">Список подтверждаемых документов</param>
        /// <param name="rejectedDocuments">Список документов, отправляемых на этап исключения</param>
        public void Process(int batchId, int taskId, IEnumerable<Document> acceptedDocuments, IEnumerable<Document> rejectedDocuments)
        {
            if (!_client.OpenTask(_sessionId, taskId))
            {
                MessageBox.Show("Task is locked");
                return;
            }
            var ids = new FlexiCapture.docIds();
            foreach (var rejectedDocument in rejectedDocuments)
            {
                ids.Add(rejectedDocument.Id);
            }
            if (ids.Count > 0)
            {
                _client.CreateTask(_sessionId, batchId, _exceptionsStageId, 0, RejectReason, 0, ids, false);
            }
            _client.OpenBatch(_sessionId, batchId);
            var hasDocuments = false;
            try
            {
                var uri = _client.Endpoint.Address.Uri.OriginalString.Replace("/API/v1/Soap", "/FileService/v1");
                foreach (var document in acceptedDocuments)
                {
                    hasDocuments = true;

                    var buffer = new byte[SignatureSize];
                    _rng.GetBytes(buffer);
                    _client.SaveAttachment(_sessionId, DocumentAttachment, document.Id, batchId, _projectId, new FlexiCapture.File
                    {
                        Bytes = buffer,
                        Name = SignatureFileName
                    });
                    document.CreateStamp();
                    UploadTextFileToServer(new Uri(uri),
                        _client.ClientCredentials.UserName.UserName,
                        _client.ClientCredentials.UserName.Password,
                        document.GetXml(), 0, _sessionId, _projectId,
                        batchId, 0, document.Id, 0, "DocumentBody"
                    ).Wait();
                }
            }
            finally
            {
                _client.CloseBatch(_sessionId, batchId);
            }
            _client.CloseTask(_sessionId, taskId, hasDocuments ? CommitUserStageType : CloseUserStageType);
        }


        private static async Task<string> UploadTextFileToServer(Uri server, string login, string password,
            string stream, int objectType, int sessionId, int projectId, int batchId, int parentId,
            int objectId, int version, string streamName)
        {
            var creds = CredentialCache.DefaultNetworkCredentials;
            if (!string.IsNullOrEmpty(login)) creds = new NetworkCredential(login, password);
            using (var handler = new HttpClientHandler { Credentials = creds })
            using (var client = new HttpClient(handler))
            {
                var encodedStreamName = Convert.ToBase64String(Encoding.Unicode.GetBytes(streamName));
                var content = new MultipartFormDataContent
                {
                    {new StringContent("Save"), "Action"},
                    {new StringContent(objectType.ToString()), "objectType"},
                    {new StringContent(sessionId.ToString()), "sessionId"},
                    {new StringContent(projectId.ToString()), "projectId"},
                    {new StringContent(batchId.ToString()), "batchId"},
                    {new StringContent(parentId.ToString()), "parentId"},
                    {new StringContent(objectId.ToString()), "objectId"},
                    {new StringContent(version.ToString()), "version"},
                    {new StringContent(encodedStreamName), "streamName"},
                    {new ByteArrayContent(Encoding.Unicode.GetBytes(stream)), "blob", "data.txt"}
                };
                var msg = client.PostAsync(server, content).Result;
                return await msg.Content.ReadAsStringAsync();
            }
        }


        /// <summary>
        /// Разрушение коннектора
        /// </summary>
        public void Dispose()
        {
            // Благодаря Interlocked только один поток сможет приступить к очистке
            if (Interlocked.Exchange(ref _disposed, 1) != 0) return;
            // Первая строка гарантирует единственный поток очистки - дальше можно не синхронизироваться
            if (_projectId != 0)
            {
                _client.CloseProject(_sessionId, _projectId);
            }
            if (_sessionId != 0)
            {
                _client.CloseSession(_sessionId);
            }
        }
    }
}
