﻿using System;
using System.ComponentModel;
using System.Linq;
using System.Threading;
using System.Windows;
using UserStageExample.FlexiService;
using UserStageExample.ViewModel;

namespace UserStageExample
{
    /// <summary>
    /// Обработчик окна
    /// </summary>
    public partial class MainWindow : Window
    {
        /// <summary>
        /// Главный контроллер
        /// </summary>
        private readonly ServiceViewModel _model = null;

        /// <summary>
        /// Коннектор FlexiCapture
        /// </summary>
        private FlexiConnectorService _service = null;

        /// <summary>
        /// Факт разрушения окна
        /// </summary>
        private int _disposing = 0;

        /// <summary>
        /// Создаёт новый обработчик окна
        /// </summary>
        public MainWindow()
        {
            _model = new ServiceViewModel();
            DataContext = _model;
            AppDomain.CurrentDomain.UnhandledException += Unhandled;
            InitializeComponent();
        }

        private void Unhandled(object sender, UnhandledExceptionEventArgs e)
        {
            if (_service != null)
            {
                _service.Dispose();
            }
        }

        /// <summary>
        /// Вызывается в момент закрытия окна
        /// </summary>
        /// <param name="e">Аргументы</param>
        protected override void OnClosing(CancelEventArgs e)
        {
            base.OnClosing(e);
            if (Interlocked.Exchange(ref _disposing, 1) == 0)
            {
                if (_service != null) _service.Dispose();
            }
        }

        /// <summary>
        /// Асинхронный поток, взаимодействующий с FlexiCapture
        /// </summary>
        /// <param name="state">Состояние</param>
        private void ExtractTasks(object state)
        {
            try
            {
                _model.WaitLoading();
                _service = new FlexiConnectorService(
                    new Uri(_model.Auth.ServerUri), 
                    _model.Auth.AuthType == AuthType.Windows, 
                    _model.Auth.Login, 
                    PasswordBox.Password);
            }
            catch (Exception e)
            {
                _model.RevertAuth(e.Message);
                return;
            }
            _model.Auth.Save();
            while (_disposing == 0)
            {
                var tasks = _service.GetTasks();
                Dispatcher.Invoke(() => _model.UpdateTasks(tasks));
                Thread.Sleep(2000);
            }
        }

        /// <summary>
        /// Подтверждает документ
        /// </summary>
        /// <param name="sender">кнопка, отправившая сообщение</param>
        /// <param name="e">Аргументы</param>
        private void Accept(object sender, RoutedEventArgs e)
        {
            var element = sender as FrameworkElement;
            if (element == null) return;
            var doc = element.DataContext as DocumentViewModel;
            if (doc == null) return;
            if (!_model.Accept(doc)) return;
            var task = doc.Task;
            _service.Process(task.BatchId, task.Id,
                task.Documents.Where(x => x.IsAccepted).Select(x => x.Document),
                task.Documents.Where(x => x.IsRejected).Select(x => x.Document));
        }

        /// <summary>
        /// Отклоняет документ
        /// </summary>
        /// <param name="sender">кнопка, отправившая сообщение</param>
        /// <param name="e">Аргументы</param>
        private void Reject(object sender, RoutedEventArgs e)
        {
            var element = sender as FrameworkElement;
            if (element == null) return;
            var doc = element.DataContext as DocumentViewModel;
            if (doc == null) return;
            if (!_model.Reject(doc)) return;
            var task = doc.Task;
            _service.Process(task.BatchId, task.Id,
                task.Documents.Where(x => x.IsAccepted).Select(x => x.Document),
                task.Documents.Where(x => x.IsRejected).Select(x => x.Document));
        }

        /// <summary>
        /// Производит авторизацию
        /// </summary>
        /// <param name="sender">кнопка, отправившая сообщение</param>
        /// <param name="e">Аргументы</param>
        private void Login(object sender, RoutedEventArgs e)
        {
            ThreadPool.QueueUserWorkItem(ExtractTasks);
        }

        /// <summary>
        /// Выходит из приложения
        /// </summary>
        /// <param name="sender">кнопка, отправившая сообщение</param>
        /// <param name="e">Аргументы</param>
        private void Exit(object sender, RoutedEventArgs e)
        {
            Close();
        }
    }
}
