﻿using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.ComponentModel;
using System.Runtime.CompilerServices;
using UserStageExample.FlexiService;

namespace UserStageExample.ViewModel
{
    /// <summary>
    /// Режим главного окна
    /// </summary>
    internal enum WindowMode
    {
        /// <summary>
        /// Форма входа в систему
        /// </summary>
        Auth,
        /// <summary>
        /// Плита ожидания
        /// </summary>
        Waiting,
        /// <summary>
        /// Общий вид, отображающий задачи
        /// </summary>
        Loaded
    }

    /// <summary>
    /// Основной контроллер приложения
    /// </summary>
    internal class ServiceViewModel : INotifyPropertyChanged
    {
        /// <summary>
        /// Текущий режим главного окна
        /// </summary>
        private WindowMode _mode = WindowMode.Auth;

        /// <summary>
        /// Отдельный контроллер авторизации
        /// </summary>
        private AuthViewModel _auth = new AuthViewModel();

        /// <summary>
        /// Оповещение о смене одного из свойств контроллера
        /// </summary>
        public event PropertyChangedEventHandler PropertyChanged;

        /// <summary>
        /// Задержка обновления
        /// </summary>
        private DateTime _updateDelay = DateTime.MinValue;

        /// <summary>
        /// Создаёт новый контроллер приложения
        /// </summary>
        public ServiceViewModel()
        {
            Tasks = new ObservableCollection<TaskViewModel>();
        }

        /// <summary>
        /// Списко обрабатываемых задач
        /// </summary>
        public ObservableCollection<TaskViewModel> Tasks { get; private set; }

        /// <summary>
        /// Контроллер аутентификации
        /// </summary>
        public AuthViewModel Auth
        {
            get { return _auth; }
            private set { _auth = value; OnPropertyChanged(); }
        }

        /// <summary>
        /// Режим работы главного окна
        /// </summary>
        public WindowMode Mode
        {
            get { return _mode; }
            private set { _mode = value; OnPropertyChanged(); }
        }

        /// <summary>
        /// Вызывает событие PropertyChanged
        /// </summary>
        /// <param name="propertyName">Имя изменившегося свойства</param>
        protected virtual void OnPropertyChanged([CallerMemberName] string propertyName = null)
        {
            var h = PropertyChanged;
            if (h != null) h.Invoke(this, new PropertyChangedEventArgs(propertyName));
        }

        /// <summary>
        /// Обновляет набор задач, основываясь на данных сервера
        /// </summary>
        /// <param name="tasks">Список задач, полученных с сервера</param>
        public void UpdateTasks(Dictionary<int, Task> tasks)
        {
            // Если обновление задеражно, то новые задачи игнорируются
            if (_updateDelay > DateTime.Now) return;
            // Обновление существующих тасков и удаление исчезнувших
            for (var i = 0; i < Tasks.Count; i++)
            {
                var taskViewModel = Tasks[i];
                Task task;
                if (tasks.TryGetValue(taskViewModel.Id, out task))
                {
                    tasks.Remove(taskViewModel.Id);
                    taskViewModel.Update(task);
                }
                else
                {
                    Tasks.RemoveAt(i--);
                }
            }
            // Добавление новых тасков
            foreach (var task in tasks)
            {
                Tasks.Add(new TaskViewModel(task.Key, task.Value));
            }
            Mode = WindowMode.Loaded;
        }

        /// <summary>
        /// Отклоняет документ
        /// </summary>
        /// <param name="document">Отклоняемый документ</param>
        /// <returns>Можно ли отправлять задачу</returns>
        public bool Reject(DocumentViewModel document)
        {
            document.Reject();
            return UpdateTask(document.Task);
        }

        /// <summary>
        /// Подтверждает документ
        /// </summary>
        /// <param name="document">Подтверждаемый документ</param>
        /// <returns>Можно ли отправлять задачу</returns>
        public bool Accept(DocumentViewModel document)
        {
            document.Accept();
            return UpdateTask(document.Task);
        }

        /// <summary>
        /// Обновляет задачу
        /// </summary>
        /// <param name="task">Обновляемая задача</param>
        /// <returns>Можно ли отправлять задачу</returns>
        private bool UpdateTask(TaskViewModel task)
        {
            if (task.HasUntagged) return false;
            Tasks.Remove(task);
            // Обновление формочки останавливается на пару секунд для того
            // чтобы закрыая задача не появилась вновь.
            _updateDelay = DateTime.Now.AddSeconds(3);
            return true;
        }

        /// <summary>
        /// Переключает окно в режим ожидания
        /// </summary>
        public void WaitLoading()
        {
            Mode = WindowMode.Waiting;
        }

        /// <summary>
        /// Возвращает окно в режим аутентификации, устанавливает сообщение об ошибке
        /// </summary>
        /// <param name="message">Сообщение об ошибке</param>
        public void RevertAuth(string message)
        {
            Auth.LoginResult = message;
            Mode = WindowMode.Auth;
        }
    }
}
